#pragma once

#include "dm_lnx_Protocol.h"
#include "HashTable.h"
#include "lkmUtils.h"

typedef struct PendingRead   PendingRead;
typedef struct PendingNotify PendingNotify;
typedef enum ConnectionFlag  ConnectionFlag;
typedef struct Connection    Connection;
typedef struct Hook          Hook;
typedef struct MemBlock      MemBlock;

//..............................................................................

struct PendingRead
{
	struct list_head m_link;
	wait_queue_head_t m_waitQueue;
	void* m_buffer; // kernel memory
	size_t m_size;
	ssize_t m_result;
	bool m_isCompleted;
};

// . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

struct PendingNotify
{
	struct list_head m_link;

	dm_NotifyHdr m_notifyHdr;

	// followed by notification-specific data
};

// . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

struct Connection
{
	Hook* m_hook;
	struct list_head m_hookLink;
	struct inode* m_inode;
	const char* m_path;
	const dm_IoctlDesc* m_ioctlDescArray;
	HashTable m_ioctlDescMap;

	struct mutex m_lock;
	struct file* m_originalFilp;
	wait_queue_head_t m_notificationWaitQueue;
	struct list_head m_pendingReadList;
	struct list_head m_pendingNotifyList;
	size_t m_pendingReadCount;
	size_t m_pendingNotifyCount;
	size_t m_pendingNotifySize;
	size_t m_pendingNotifySizeLimit;

	long m_refCount;
	uint m_fileFlags;
};

// . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

int
Connection_create (
	Connection** connection,
	Hook* Hook,
	struct file* filp,
	const dm_IoctlDesc __user* ioctlDescArray_u,
	size_t ioctlDescCount,
	size_t pendingNotifySizeLimit,
	uint fileFlags // mostly curious about O_NONBLOCK
	);

static
inline
void
Connection_addRef (Connection* self)
{
	atomicInc (&self->m_refCount);
}

long
Connection_release (Connection* self);

void
Connection_disconnect (Connection* self);

void
Connection_detachFromHook (Connection* self);

ssize_t
Connection_read (
	Connection* self,
	void __user* buffer_u,
	size_t size
	);

void
Connection_notify (
	Connection* self,
	struct file* filp,
	uint16_t code,
	int result,
	uint32_t pid,
	uint32_t tid,
	uint64_t timestamp,
	MemBlock* paramBlockArray,
	size_t paramBlockCount
	);

bool
Connection_p_preIoctlNotify (
	Connection* self,
	MemBlock* paramBlockArray,
	size_t paramBlockCount
	);

//..............................................................................
